// gcc -Wall -O3 try_sum.c -lgmp -L/usr/local/lib -lmpfr -o try_sum
// mcopy -o -v try_sum.c a:

#include <stdio.h>
#include <mpfr.h>

void sum_with_bound(mpfr_t s, mpfr_t e, mpfr_t x[], unsigned int n) {
  // N. Higham "Accuracy and stability of numerical algorithms" exercise 4.3
  // returns s and e s.t. |s-sum(x[i],i=0..n-1)|<e
  // s and e are computed to their initialized precision
  int i;
  mpfr_t u,g,t;
  mpfr_set_ui(s,0,GMP_RNDN);
  if (n==0) {
    mpfr_set_ui(e,0,GMP_RNDN);
    return;
  } 
  for (i=0; i<n; i++) mpfr_add(s,s,x[i],GMP_RNDN);
  mpfr_init(u);
  mpfr_init(g);
  mpfr_init(t);
  mpfr_set_si_2exp(u,1,-mpfr_get_prec(s),GMP_RNDU); // unit roundoff
  mpfr_mul_ui(t,u,n-2,GMP_RNDU);
  mpfr_set(g,t,GMP_RNDU);
  mpfr_ui_sub(t,1,t,GMP_RNDD);
  mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-2}
  mpfr_abs(t,x[0],GMP_RNDU);
  mpfr_mul(t,t,g,GMP_RNDU);
  mpfr_set(e,t,GMP_RNDU);
  if (n>1) {
    mpfr_abs(t,x[1],GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(e,e,t,GMP_RNDU);
  }
  for (i=2; i<n; i++) {
    mpfr_mul_ui(t,u,n-i,GMP_RNDU);
    mpfr_set(g,t,GMP_RNDU);
    mpfr_ui_sub(t,1,t,GMP_RNDD);
    mpfr_div(g,g,t,GMP_RNDU); // gamma_{n-i}
    mpfr_abs(t,x[i],GMP_RNDU);
    mpfr_mul(t,t,g,GMP_RNDU);
    mpfr_add(e,e,t,GMP_RNDU);
  }
  mpfr_clear(u);
  mpfr_clear(g);
  mpfr_clear(t);
}

void test0() {
  // evalf[100](add(sin(i),i=0..9));
  const int n=10,bits=100;
  int i;
  printf("test0\n");
  mpfr_t e,s,z,x[n];
  mpfr_init2(s,bits);
  mpfr_init2(z,300);
  mpfr_set_str(z,"1.955209482107380269033313299261795324452547619149472349814341481069843043218182727515900772201747674",10,GMP_RNDN);
  mpfr_init2(e,64);
  for (i=0; i<n; i++) mpfr_init2(x[i],bits);
  for (i=0; i<n; i++) mpfr_set_si(x[i],i,GMP_RNDN);
  for (i=0; i<n; i++) mpfr_sin(x[i],x[i],GMP_RNDN);
  sum_with_bound(s,e,x,n);
  printf("sum=")+mpfr_out_str(stdout,10,bits/3,s,GMP_RNDN)+printf("\n");
  printf("error bound =")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  mpfr_sub(e,z,s,GMP_RNDN);
  printf("actual error=")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  // FIXME clear
}

void test1() {
  // > evalf[100](add(1/i,i=1..1000));
  const int n=1000,bits=100;
  int i;
  mpfr_t e,s,z,x[n];
  printf("test1\n");
  mpfr_init2(s,bits);
  mpfr_init2(z,300);
  mpfr_set_str(z,"7.485470860550344912656518204333900176521679169708803665773626749957699349165202440959934437411845081",10,GMP_RNDN);
  mpfr_init2(e,64);
  for (i=0; i<n; i++) mpfr_init2(x[i],bits);
  for (i=0; i<n; i++) mpfr_set_si(x[i],i+1,GMP_RNDN);
  for (i=0; i<n; i++) mpfr_ui_div(x[i],1,x[i],GMP_RNDN);
  sum_with_bound(s,e,x,n);
  printf("sum=")+mpfr_out_str(stdout,10,bits/3,s,GMP_RNDN)+printf("\n");
  printf("error bound =")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  mpfr_sub(e,z,s,GMP_RNDN);
  printf("actual error=")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  // FIXME clear
}

void test2() {
  // evalf[1000](add(1/i^2,i=1..100000));
  const int n=100000,bits=100;
  long i;
  mpfr_t e,s,z,x[n];
  printf("test2\n");
  mpfr_init2(s,bits);
  mpfr_init2(z,1000);
  mpfr_set_str(z,"1.6449240668982262698057485033126918556475213298115603424898872337015429349374987426322069141742813721268751881465063581964734108090055328816463859207358377357448214140227529335929277925015034079959440512476188758093226500837227203471787848666015074439347618549586151372588037133068485936",10,GMP_RNDN);
  mpfr_init2(e,64);
  for (i=0; i<n; i++) mpfr_init2(x[i],bits);
  for (i=0; i<n; i++) mpfr_set_si(x[i],(i+1)*(i+1),GMP_RNDN);
  for (i=0; i<n; i++) mpfr_ui_div(x[i],1,x[i],GMP_RNDN);
  sum_with_bound(s,e,x,n);
  printf("sum=")+mpfr_out_str(stdout,10,bits/3,s,GMP_RNDN)+printf("\n");
  printf("error bound =")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  mpfr_sub(e,z,s,GMP_RNDN);
  printf("actual error=")+mpfr_out_str(stdout,10,10,e,GMP_RNDN)+printf("\n");
  // FIXME clear
}

int main() {
  test0();
  test1();
  test2();
  return 0;
}
